/***************************************************************************** 
P Lambda List.

LAMBDA EDITION.
by XavSnap 12/2018

FROM
TapList - ZX Spectrum tape file to ASCII text converter 

Copyright (c)2002 Chris Cowley.  <cco...@grok.co.uk> 
Microsoft Visual Studio 2012 Port by Volker Bartheld. <w e b 2 0 1 8 AT b a r t h e l d . n e t> 

This program is free software; you can redistribute it and/or 
modify it under the terms of the GNU General Public License 
as published by the Free Software Foundation; either version 2 
of the License, or (at your option) any later version. 
This program is distributed in the hope that it will be useful, 
but WITHOUT ANY WARRANTY; without even the implied warranty of 
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
GNU General Public License for more details. 

You should have received a copy of the GNU General Public License 
along with this program; if not, write to the Free Software 
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. 

*****************************************************************************/ 

#include <stdio.h> 
#include <string.h> 


#define PROG_VER "1.00" 

int DeTokenize(const unsigned char *In, int LineLen, char *Out); 
void ConCat(char *Out, int *Pos, const char *Text); 


int main(int argc, char* argv[]) 
{ 
  FILE *fIn; 
  unsigned short ProgAddr, LineNum, LineLen; 
  unsigned char LineData[65535]; 
  char LineText[65535]; 

  if (argc != 2) 
  { 
    printf("PlambdaList v"); 
    printf(PROG_VER);  
        printf("\nBy XavSnap 2018 [Chris Cowley (c)2002;Volker Bartheld update 2018]\n"); 
    printf("\n\nUsage: %s filename.sna > outputfile.txt\n\n",argv[0]); 
    return(1); 
  } 

  if( (fIn = fopen(argv[1], "rb" )) == NULL ) 
  { 
    fprintf(stderr,"ERROR: Unable to open file %s\n",argv[1]); 
    return(2); 
  } 

  /* All 48K snapshots are 49179 bytes, check to ensure our .sna is at least this long */ 
  fseek(fIn, SEEK_SET, SEEK_END); 

  /* 
  ** Process the snapshot pointed to by fIn ** 
  */ 

   // Offset is: PROG=23635-16384+27 

  const int PROG_ADDR_OFFS=900; 

//  fread(LineData,1,2,fIn); 
//  ProgAddr = LineData[0]+LineData[1]*256; 

ProgAddr=PROG_ADDR_OFFS;
 while(  LineData[0] != 0x76 ) 
  { 
  fseek(fIn,ProgAddr,SEEK_SET); 
  fread(LineData,1,1,fIn);
//    printf(">>%d\n",LineData[0]); 
  ProgAddr++;
} 
  ProgAddr++;

  /* seek to PROG area */ 
 // const int PROG_OFFS=ProgAddr; 
 // fseek(fIn,PROG_OFFS,SEEK_SET); 

  while( !feof(fIn) ) 
  { 
    fread(LineData,1,2,fIn); 
    LineNum = LineData[0]*256+LineData[1]; 

    if (LineNum > 9999) break; 

    fread(LineData,1,2,fIn); 
    LineLen = LineData[0]+LineData[1]*256; 

    fread(LineData,1,LineLen,fIn); 
    LineData[LineLen]=0; /* Terminate the line data */ 

    DeTokenize(LineData,LineLen,LineText); 

    printf("%d %s\n",LineNum,LineText); 
  } 

  fclose(fIn); 

  return(0); 
} 


/***************************************************************************** 
DeTokenize() 

'In' should point to a string containing tokenized ZX Spectrum text. 
'LineLen' should contain the length of the Input data in characters. 
The expanded Text will be placed in 'Out', which must be large enough 
to contain the full text. 
*****************************************************************************/ 
int DeTokenize(const unsigned char *In, int LineLen, char *Out) 
{ 
  int i = 0, o = 0; 
  
  for(i=0;i<LineLen;i++) 
  { 
		    switch (In[i]) 
		    { 
		    case 0: 
			   ConCat(Out,&o," ");		
		       break; 
			case 1: 
			   ConCat(Out,&o,"\\' ");		
		       break; 
			case 2: 
			   ConCat(Out,&o,"\\ '");		
		       break; 
			case 3: 
			   ConCat(Out,&o,"\\''");		
		       break; 
			case 4: 
		 	   ConCat(Out,&o,"\\. ");		
		       break; 
			case 5: 
			   ConCat(Out,&o,"\\: ");		
		       break; 
			case 6: 
			   ConCat(Out,&o,"\\.'");		
		       break; 
			case 7: 
			   ConCat(Out,&o,"\\:'");		
		       break; 
			case 8: 
			   ConCat(Out,&o,"\\##");		
		       break;
		     
			case 9: 
			   ConCat(Out,&o,"\\:>");		
		       break; 
			case 10: 
			   ConCat(Out,&o,"\\<:");		
		       break; 
		       
			case 126: 
			   i=i+5;
		       break;	
			   	     
			case 128: 
			   ConCat(Out,&o,"\\::");		
		       break;
			case 129: 
			   ConCat(Out,&o,"\\.:");		
		       break;
			case 130: 
			   ConCat(Out,&o,"\\:.");		
		       break;
			case 131: 
			   ConCat(Out,&o,"\\..");		
		       break;
			case 132: 
			   ConCat(Out,&o,"\\':'");		
		       break;
			case 133: 
			   ConCat(Out,&o,"\\ :");		
		       break;
			case 134: 
			   ConCat(Out,&o,"\\'.");		
		       break;
			case 135: 
			   ConCat(Out,&o,"\\ .");		
		       break;
			case 136: 
			   ConCat(Out,&o,"\\@@");		
		       break;
		       
			case 137: 
			   ConCat(Out,&o,"\\>:");		
		       break;
			case 138: 
			   ConCat(Out,&o,"\\:<");		
		       break;
	
//-------------------------------------------
	       
			case 11: 
			   ConCat(Out,&o,"\"");		
		       break;
			case 12: 
			   ConCat(Out,&o,"");		
		       break;
			case 13: 
			   ConCat(Out,&o,"$");
		       break;
		    case 14: 
			   ConCat(Out,&o,":");
		       break; 
			case 15: 
			   ConCat(Out,&o,"?");		
		       break;
		    case 16: 
			    ConCat(Out,&o,"(");
		      break; 
		    case 17: 
			    ConCat(Out,&o,")");
    	      break;
		    case 18: 
		        ConCat(Out,&o,">");
		      break; 
		    case 19: 
			    ConCat(Out,&o,"<");
		      break; 
		    case 20: 
		        ConCat(Out,&o,"=");
		      break; 
		    case 21: 
		        ConCat(Out,&o,"+");
		      break; 
		    case 22: 
			    ConCat(Out,&o,"-");
		      break; 
 		    case 23: 
			    ConCat(Out,&o,"*");
		      break; 
 		    case 24: 
			    ConCat(Out,&o,"/");
		      break; 
 		    case 25: 
			    ConCat(Out,&o,";");
		      break; 
 		    case 26: 
			    ConCat(Out,&o,",");
		      break; 
 		    case 27: 
			    ConCat(Out,&o,".");
		      break; 
 		    case 28: 
			    ConCat(Out,&o,"0");
		      break; 
		    case 29: 
			    ConCat(Out,&o,"1");
		      break; 
		    case 30: 
			    ConCat(Out,&o,"2");
		      break; 
		    case 31: 
			    ConCat(Out,&o,"3");
		      break; 
		    case 32: 
			    ConCat(Out,&o,"4");
		      break; 
		    case 33: 
			    ConCat(Out,&o,"5");
		      break; 
		    case 34: 
			    ConCat(Out,&o,"6");
		      break; 
		    case 35: 
			    ConCat(Out,&o,"7");
		      break; 
		    case 36: 
			    ConCat(Out,&o,"8");
		      break; 
		    case 37: 
			    ConCat(Out,&o,"9");
		      break; 
 		      
// ---------------------------------------------
	      
		    case 64: 
		      ConCat(Out,&o," THEN "); 
		      break; 		    
			case 65: 
		      ConCat(Out,&o," TO "); 
		      break; 		    
			case 66: 
		      ConCat(Out,&o," STEP "); 
		      break; 		    
			case 67: 
		      ConCat(Out,&o,"RND"); 
		      break; 
		    case 68: 
		      ConCat(Out,&o," INKEY$"); 
		      break; 
		    case 69: 
		      ConCat(Out,&o,"PI");
		      break; 
		    case 70: 
		      ConCat(Out,&o,"INK "); 
		      break; 
		    case 71: 
		      ConCat(Out,&o,"PAPER "); 
		      break; 
			case 72: 
		      ConCat(Out,&o,"BORDER "); 
		      break; 	      
		      		      
// ---------------------------------------------
		    case 192: 
		      ConCat(Out,&o,"CODE "); 
		      break; 
		    case 193: 
		      ConCat(Out,&o,"VAL "); 
		      break; 
		    case 194: 
		      ConCat(Out,&o,"LEN "); 
		      break; 
		    case 195: 
		      ConCat(Out,&o,"SIN "); 
		      break; 
		    case 196: 
		      ConCat(Out,&o,"COS "); 
		      break; 
		    case 197: 
		      ConCat(Out,&o,"TAN "); 
		      break; 
		    case 198: 
		      ConCat(Out,&o,"ASN "); 
		      break; 
		    case 199: 
		      ConCat(Out,&o,"ACS "); 
		      break; 
		    case 200: 
		      ConCat(Out,&o,"ATN"); 
		      break; 
		    case 201: 
		      ConCat(Out,&o,"LOG"); 
		      break; 
		    case 202: 
		      ConCat(Out,&o,"EXP"); 
		      break; 
		    case 203: 
		      ConCat(Out,&o,"INT"); 
		      break; 
		    case 204: 
		      ConCat(Out,&o,"SQR"); 
		      break; 
		    case 205: 
		      ConCat(Out,&o,"SGN"); 
		      break; 
		    case 206: 
		      ConCat(Out,&o,"ABS"); 
		      break; 
		    case 207: 
		      ConCat(Out,&o,"PEEK "); 
		      break; 
		    case 208: 
		      ConCat(Out,&o," USR "); 
		      break; 
		    case 209: 
		      ConCat(Out,&o," STR$"); 
		      break; 
		    case 210: 
		      ConCat(Out,&o," CHR$"); 
		      break; 
		    case 211: 
		      ConCat(Out,&o,"NOT "); 
		      break; 
		    case 212: 
		      ConCat(Out,&o," AT "); 
		      break; 
		    case 213: 
		      ConCat(Out,&o,"TAB"); 
		      break; 
		    case 214: 
		      ConCat(Out,&o,"**"); 
		      break; 
		    case 215: 
		      ConCat(Out,&o," OR "); 
		      break; 
		    case 216: 
		      ConCat(Out,&o," AND "); 
		      break; 
		    case 217: 
		      ConCat(Out,&o,"<="); 
		      break; 
		    case 218: 
		      ConCat(Out,&o,">="); 
		      break; 
		    case 219: 
		      ConCat(Out,&o,"<>"); 
		      break; 
		    case 220: 
		      ConCat(Out,&o,"TEMPO "); 
		      break; 
		    case 221: 
		      ConCat(Out,&o,"MUSIC "); 
		      break; 
		    case 222: 
		      ConCat(Out,&o,"SOUND "); 
		      break; 
		    case 223: 
		      ConCat(Out,&o,"BEEP "); 
		      break; 
		    case 224: 
		      ConCat(Out,&o,"NOBEEP "); 
		      break; 
		    case 225: 
		      ConCat(Out,&o,"LPRINT "); 
		      break; 
		    case 226: 
		      ConCat(Out,&o,"LLIST "); 
		      break; 
		    case 227: 
		      ConCat(Out,&o,"STOP "); 
		      break; 
		    case 228: 
		      ConCat(Out,&o,"SLOW "); 
		      break; 
		    case 229: 
		      ConCat(Out,&o,"FAST "); 
		      break; 
		    case 230: 
		      ConCat(Out,&o,"NEW "); 
		      break; 
		    case 231: 
		      ConCat(Out,&o,"SCROLL "); 
		      break; 
		    case 232: 
		      ConCat(Out,&o,"CONT "); 
		      break; 
		    case 233: 
		      ConCat(Out,&o,"DIM "); 
		      break; 
		    case 234: 
		      ConCat(Out,&o,"REM "); 
		      break; 
		    case 235: 
		      ConCat(Out,&o,"FOR "); 
		      break; 
		    case 236: 
		      ConCat(Out,&o,"GOTO "); 
		      break; 
		    case 237: 
		      ConCat(Out,&o,"GOSUB "); 
		      break; 
		    case 238: 
		      ConCat(Out,&o,"INPUT "); 
		      break; 
		    case 239: 
		      ConCat(Out,&o,"LOAD "); 
		      break; 
		    case 240: 
		      ConCat(Out,&o,"LIST "); 
		      break; 
		    case 241: 
		      ConCat(Out,&o,"LET "); 
		      break; 
		    case 242: 
		      ConCat(Out,&o,"PAUSE "); 
		      break; 
		    case 243: 
		      ConCat(Out,&o,"NEXT "); 
		      break; 
		    case 244: 
		      ConCat(Out,&o,"POKE "); 
		      break; 
		    case 245: 
		      ConCat(Out,&o,"PRINT "); 
		      break; 
		    case 246: 
		      ConCat(Out,&o,"PLOT "); 
		      break; 
		    case 247: 
		      ConCat(Out,&o,"RUN "); 
		      break; 
		    case 248: 
		      ConCat(Out,&o,"SAVE "); 
		      break; 
		    case 249: 
		      ConCat(Out,&o,"RAND "); 
		      break; 
		    case 250: 
		      ConCat(Out,&o,"IF "); 
		      break; 
		    case 251: 
		      ConCat(Out,&o,"CLS "); 
		      break; 
		    case 252: 
		      ConCat(Out,&o,"UNPLOT "); 
		      break; 
		    case 253: 
		      ConCat(Out,&o,"CLEAR "); 
		      break; 
		    case 254: 
		      ConCat(Out,&o,"RETURN "); 
		      break; 
		    case 255: 
		      ConCat(Out,&o,"COPY "); 
		      break;
		
		    default: 
		      if (In[i]>=38 && In[i]<=63) Out[o++] = In[i]+27;  /* A to Z */
		      if (In[i]>=166 && In[i]<=191) Out[o++] = In[i]-128+59;   /*  a to z */
		      
		      if (In[i]>=28 && In[i]<=37) Out[o++] = In[i]+20;  /* 0 to 9 */
		      if (In[i]>=156 && In[i]<=165) Out[o++] = In[i]-128+20;   /* inv 0 to 9 */
			  	      
               
		    } 
		    
  } 

  Out[o]=0; 
  return strlen(Out); 
  
} 


/***************************************************************************** 
ConCat() 

Places the null-terminated string pointed to by 'Text' at position 'Pos' 
in 'Out'. 'Out' must be large enough to contain the newly placed text. 
*****************************************************************************/ 
void ConCat(char *Out, int *Pos, const char *Text) 
{ 
  int n = 0; 

  /* Stop two adjacent spaces appearing */ 
  if ( (*Pos > 0) && (Out[*Pos-1] == 32) && (Text[0] == 32) ) (*Pos)--; 

  while(Text[n] != 0) 
  { 
    Out[*Pos+n] = Text[n]; 
    n++; 
  } 

  *Pos += n; 
} 
